from datetime import date

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models import Prefetch
from django.db.models import Q
from django.db.transaction import atomic

from concentrator.change import ChangeSource
from concentrator.change import StorageHelper
from concentrator.models import DocExtraInfo
from concentrator.smev3_v321.base.constants import CODE_ERROR
from concentrator.smev3_v321.base.constants import CODE_OK
from concentrator.smev3_v321.order.constants import DECLARATION_CHANGED
from concentrator.smev3_v321.order.tasks import UpdateOrderRequestSendingTask
from kinder import logger
from kinder.core.children.models import Children
from kinder.core.children.models import ChildrenDelegate
from kinder.core.children.models import Delegate
from kinder.core.children.models import DelegateTypeEnumerate
from kinder.core.children.models import GenderEnumerate
from kinder.core.declaration.enum import DeclarationTypeInteractionEnum as DTIE
from kinder.core.declaration.enum import DeclPortalEnum
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationPrivilege
from kinder.core.declaration.models import DeclarationSourceEnum
from kinder.core.declaration.models import DeclarationUnit
from kinder.core.declaration_status.models import DSS
from kinder.core.declaration_status.models import DeclarationStatus
from kinder.core.dict.models import DULDelegateType
from kinder.core.dict.models import GroupOrientationDocuments
from kinder.core.dict.models import GroupSpec
from kinder.core.dict.models import GroupType
from kinder.core.dict.models import HealthNeed
from kinder.core.dict.models import WorkType
from kinder.core.group.models import Pupil
from kinder.core.privilege.models import Privilege
from kinder.core.unit.models import Unit
from kinder.core.utils.fias import get_fias_code
from kinder.plugins.contingent.models import DelegateContingent
from kinder.webservice.api.declaration import ApiException
from kinder.webservice.api.declaration import get_declaration_by_client_id
from ..constants import CHANGES_DATA_TEMPATE
from ..constants import DECLARATION_RECEIVED_CODE
from ..constants import DECLARATION_RECEIVED_COMMENT

from ..models import OrderRequest
from .changes import Smev3ChangeHelper
from .changes import Smev3ChangesMap
from .changes import Smev3DeclarationDocsChangeHelper
from .changes import Smev3DeclarationPrivilegeChangeHelper
from .changes import Smev3DeclarationUnitChangeHelper
from .constants import MAX_AGE
from .constants import ApplicationRequestMessage as _Message
from .constants import Response
from .utils import process_attachments
from ..constants import CHANGE_DECLARATION_CODE
from ..constants import CHANGE_DECLARATION_COMMENT
from ..helpers import push_change_order_info_request
from ..models import DeclarationOriginMessageID
from ..models import OrderRequest


class Application:
    """Сохранение заявления из концентратора СМЭВ 3."""

    def __init__(self, application, attachments):
        self._application = application
        self._attachments = attachments

    def _create_child(self):
        """Создает ребенка.

        :return: Новый ребенок
        :rtype: Children

        :raise: ValidationError
        """

        # Блок данных с общей информацией о ребенке
        child_info = self._application.ChildInfo
        # Блок данных с информацией о ДУЛ ребенка
        child_birth_doc = child_info.ChildBirthDocRF
        # Блок данных с информацией о адресе проживания ребенка
        address = self._application.Address
        # Блок данных о здоровье ребенка
        medical_report = self._application.MedicalReport

        child_params = {}

        adaptation_program = self._application.AdaptationProgram
        if adaptation_program.AdaptationGroupType:
            try:
                # Специфика группы не является уникальным значением в ЕСНСИ,
                # поэтому для однозначной идентификации выполняется поиск
                # по связке значений: код ЕСНСИ и Направленность группы.
                child_params['health_need'] = (
                    HealthNeed.objects.get(
                        esnsi_code=adaptation_program.AdaptationGroupType.code,
                        group_type__esnsi_code=(
                            adaptation_program.AdaptationGroup.code)))
            except HealthNeed.DoesNotExist:
                raise ValidationError(
                    'Указано неверное значение AdaptationGroupType')

        if adaptation_program.NeedSpecialCareConditions is not None:
            child_params['health_need_special_support'] = (
                adaptation_program.NeedSpecialCareConditions)

        if medical_report:

            try:
                group_orientation_document = (
                    GroupOrientationDocuments.objects.get(
                        esnsi_code=medical_report.DocName.code))
            except GroupOrientationDocuments.DoesNotExist:
                raise ValidationError('Указано неверное значение DocName')

            if medical_report.DocSeries:
                child_params['health_series'] = medical_report.DocSeries

            if medical_report.DocExpirationDate:
                child_params['health_need_expiration_date'] = (
                    medical_report.DocExpirationDate)

            child_params.update({
                'health_need_confirmation': f'{group_orientation_document.id}',
                'health_number': medical_report.DocNumber,
                'health_need_start_date': medical_report.DocIssueDate,
                'health_issued_by': medical_report.DocIssued
            })

        child_params = update_middlename_params(
            child_info.ChildMiddleName, child_params)

        child = (
            Children.objects.create(
                firstname=child_info.ChildName,
                surname=child_info.ChildSurname,
                patronymic=child_info.ChildMiddleName,
                date_of_birth=child_info.ChildBirthDate,
                dul_series=child_birth_doc.ChildBirthDocSeries,
                dul_number=child_birth_doc.ChildBirthDocNumber,
                dul_date=child_birth_doc.ChildBirthDocIssueDate,
                zags_act_number=child_birth_doc.ChildBirthDocActNumber,
                zags_act_date=child_birth_doc.ChildBirthDocActDate,
                zags_act_place=child_birth_doc.ChildBirthDocIssued,
                reg_address_place=get_fias_code(address.Region.code),
                reg_address_street=get_fias_code(address.Street.code),
                reg_address_full=address.FullAddress,
                reg_address_house=address.House.valueOf_,
                reg_address_house_guid=get_fias_code(address.House.code),
                reg_address_corps=address.Building1,
                reg_address_flat=address.Apartment, **child_params))

        return child

    def _create_delegate(self):
        """Создает представителя.

        :return: Новый представитель
        :rtype: Delegate

        :raise: ValidationError
        """

        # Блок данных с информацией о представителе
        person_info = self._application.PersonInfo
        person_identity_doc_info = self._application.PersonIdentityDocInfo
        address = self._application.Address

        # Доп. параметры для создания представителя
        params = {}

        try:
            params['dul_type'] = (
                DULDelegateType.objects.get(
                    esnsi_code=person_identity_doc_info.IdentityDocName.code))
        except DULDelegateType.DoesNotExist:
            raise ValidationError('Указано неверное значение IdentityDocName')

        if person_identity_doc_info.IdentityDocSeries:
            params['dul_series'] = person_identity_doc_info.IdentityDocSeries

        if person_identity_doc_info.IdentityDocIssueCode:
            params['dul_place'] = person_identity_doc_info.IdentityDocIssueCode

        params = update_middlename_params(
            person_info.PersonMiddleName, params, person_info.Parents)

        delegate = Delegate.objects.create(
            surname=person_info.PersonSurname,
            firstname=person_info.PersonName,
            patronymic=person_info.PersonMiddleName,
            phones=person_info.PersonPhone,
            email=person_info.PersonEmail,
            dul_number=person_identity_doc_info.IdentityDocNumber,
            dul_date=person_identity_doc_info.IdentityDocIssueDate,
            dul_issued_by=person_identity_doc_info.IdentityDocIssued,
            reg_address_place=get_fias_code(address.City.code),
            reg_address_street=get_fias_code(address.Street.code),
            reg_address_full=address.FullAddress,
            reg_address_house=address.House.valueOf_,
            reg_address_house_guid=get_fias_code(address.House.code),
            reg_address_flat=address.Apartment, **params)

        contingent, _ = DelegateContingent.objects.get_or_create(
            delegate=delegate)

        return delegate

    @staticmethod
    def _create_children_delegate(child, delegate):
        """Создает связь ребенка и представителя.

        :param child: Ребенок
        :type child: Children
        :param delegate: Представитель
        :type delegate: Delegate
        """
        ChildrenDelegate.objects.create(
            children=child,
            delegate=delegate)

    def _get_mo(self):
        """Возвращает МО.

        :return: МО указанных ДОО
        :rtype: Unit

        :raise: ValidationError
        """

        units = self._application.EduOrganizations.EduOrganization
        top_units = [u for u in units if u.PriorityNumber] or units
        try:
            top_unit = Unit.objects.get(id=top_units[0].code)
        except (Unit.DoesNotExist, IndexError):
            raise ValidationError('Не удалось найти МО')
        return top_unit.get_mo()

    def _create_declaration(self, child):
        """Создает заявление.

        :param child: Инстанс ребенка
        :type child: Children

        :return: Новое заявление
        :rtype: Declaration

        :raise: ValidationError
        """

        adaptation_program = self._application.AdaptationProgram
        entry_params = self._application.EntryParams

        # Доп. параметры для заявления
        params = {}

        try:
            params['desired_group_type'] = (
                GroupType.objects.get(
                    esnsi_code=adaptation_program.AdaptationGroup.code))
        except GroupType.DoesNotExist:
            raise ValidationError('Указано неверное значение AdaptationGroup')

        if adaptation_program.AgreementOnGeneralGroup is not None:
            params['consent_dev_group'] = (
                adaptation_program.AgreementOnGeneralGroup)

        if adaptation_program.AgreementOnCareGroup is not None:
            params['consent_care_group'] = (
                adaptation_program.AgreementOnCareGroup)

        try:
            params['work_type'] = (
                WorkType.objects.get(esnsi_code=entry_params.Schedule.code))
        except WorkType.DoesNotExist:
            raise ValidationError('Указано неверное значение Schedule')

        if entry_params.AgreementOnFullDayGroup is not None:
            params['consent_full_time_group'] = (
                entry_params.AgreementOnFullDayGroup)

        try:
            params['spec'] = (
                GroupSpec.objects.get(esnsi_code=entry_params.Language.code))
        except GroupSpec.DoesNotExist:
            raise ValidationError('Указано неверное значение Language')

        declaration_status = DeclarationStatus.objects.get(code=DSS.RECEIVED)

        # Дата подачи заявления
        if self._application.FilingDate:
            params['date'] = self._application.FilingDate

        return Declaration.objects.create(
            client_id=str(self._application.orderId),
            source=DeclarationSourceEnum.CONCENTRATOR,
            portal=DeclPortalEnum.PORTAL,
            type_interaction=DTIE.SMEV_3,
            children=child,
            desired_date=entry_params.EntryDate,
            status=declaration_status,
            offer_other=self._application.EduOrganizations.AllowOfferOther,
            mo=self._get_mo(), **params)

    def _add_docs(self, declaration):
        """Добавляет прикреплённые файлы к заявке."""
        if self._attachments:
            process_attachments(self._attachments, declaration)

    def _add_declaration_units(self, declaration):
        """Выполняет создание Желаемых ДОО для заявления.

        :param declaration: Заявление
        :type declaration: Declaration
        """

        # Отображение вида: идентификатор организации -> идентификатор ребенка.
        children_cache = {}

        # Разбор блока с информацией о братьях/сестрах ребенка.
        for sibling in self._application.BrotherSisterInfo:

            params = {}
            child_middle_name = sibling.ChildMiddleName
            if child_middle_name:
                params['children__patronymic__iexact'] = child_middle_name

            child_id = (
                Pupil.objects.filter(
                    grup__unit_id=sibling.EduOrganization.code,
                    children__surname__iexact=sibling.ChildSurname,
                    children__firstname__iexact=sibling.ChildName,
                    **params
                ).values_list('children_id', flat=True).first())

            if child_id:
                children_cache.setdefault(
                    sibling.EduOrganization.code, child_id)

        for unit in self._application.EduOrganizations.EduOrganization:

            params = {}

            child_id = children_cache.get(unit.code)
            if child_id:
                params['sibling_id'] = child_id

            DeclarationUnit.objects.create(
                declaration=declaration, unit_id=unit.code,
                ord=unit.PriorityNumber, **params)

    def _add_privilege(self, declaration):
        """Добавляет льготу к заявке.

        :param declaration: Заявление
        :type declaration: Declaration

        :raise: ValidationError
        """

        benefit_info = self._application.BenefitInfo

        if not benefit_info:
            return

        try:
            privilege = (
                Privilege.objects.get(
                    esnsi_code=benefit_info.BenefitCategory.code))
        except Privilege.DoesNotExist:
            raise ValidationError('Льгота не найдена')

        # Блок данных с описанием документа
        benefit_doc_info = benefit_info.BenefitDocInfo

        privilege_params = {
            'doc_date': benefit_doc_info.DocIssueDate,
            'doc_issued_by': benefit_doc_info.DocIssued
        }

        if benefit_doc_info.DocExpirationDate:
            privilege_params['_privilege_end_date'] = (
                benefit_doc_info.DocExpirationDate)

        DeclarationPrivilege.objects.create(
            declaration=declaration,
            privilege=privilege,
            **privilege_params)

    @atomic
    def create(self):
        """Создает и возвращает заявление."""

        child = self._create_child()
        delegate = self._create_delegate()
        self._create_children_delegate(child, delegate)

        declaration = self._create_declaration(child)
        self._add_privilege(declaration)
        self._add_declaration_units(declaration)
        self._add_docs(declaration)

        return declaration


class ApplicationManager:
    """Выполняет первичную валидацию, полученных данных заявления
    из концентратора СМЭВ 3, поиск существующего заявление по идентификатору
    (order_id), создание/фиксирование изменений заявления.
    """

    def __init__(self, application_type, attachments=None, message=None):
        """Конструктор.

        :param application_type: XML структура заявления
        :param attachments: Вложения
        :param message: Параметры запроса
        """

        self.application_type = application_type
        self.attachments = attachments
        self.message = message or {}

    def validate(self):
        """Выполняет первичную валидацию.

        :raise: ValidationError
        """

        today = date.today()

        entry_date = self.application_type.EntryParams.EntryDate
        birth = self.application_type.ChildInfo.ChildBirthDate

        # Валидация Желаемой даты зачисления, которая должна быть
        # равна либо больше сегодняшней даты
        if entry_date and entry_date < today:
            raise ValidationError(
                _Message.DESIRED_DATE_ERROR.format(
                    today_date=today.strftime('%d.%m.%Y')))

        # Валидация Даты рождения ребенка
        birth_border = date(today.year - MAX_AGE, today.month, today.day)
        if birth and birth <= birth_border:
            raise ValidationError(_Message.OLD)

    def get_declaration_by_order_id(self):
        """Возвращает заявление по идентификатору (order_id).

        :return: Существующего заявление
        :rtype: Optional[Declaration]
        """

        # Доп. подзапросы на получение необходимых данных для
        # последующей проверки изменений
        declaration_query = (
            Declaration.objects.select_related(
                'children'
            ).prefetch_related(
                Prefetch(
                    'children__childrendelegate_set',
                    queryset=(
                        ChildrenDelegate.objects.select_related('delegate')),
                    to_attr='childrendelegate_all')))

        try:
            declaration = get_declaration_by_client_id(
                declaration_query, self.application_type.orderId)
        except ApiException:
            declaration = None

        return declaration

    def create_declaration(self):
        """Создает новое заявление.

        :return: Ответ
        :rtype: Response
        """

        child_info = self.application_type.ChildInfo

        base_filter = (
                Q(children__surname__iexact=child_info.ChildSurname)
                & Q(children__firstname__iexact=child_info.ChildName)
                & Q(children__date_of_birth=child_info.ChildBirthDate)
                & ~Q(status__code__in=DSS.no_active_statuses())
                & Q(children__dul_series=(
                    child_info.ChildBirthDocRF.ChildBirthDocSeries))
                & Q(children__dul_number=(
                    child_info.ChildBirthDocRF.ChildBirthDocNumber))
                & Q(children__dul_date=(
                    child_info.ChildBirthDocRF.ChildBirthDocIssueDate)))

        if Declaration.objects.filter(base_filter).exists():
            # Формирует ответ с сообщением о существовании дубля в системе
            return Response(
                order_id=self.application_type.orderId,
                org_code=_Message.values[_Message.MULTIPLE],
                tech_code=CODE_ERROR, comment=_Message.MULTIPLE)

        try:
            declaration = Application(
                self.application_type, self.attachments).create()
            message_id = self.message.get('message_id')
            if message_id:
                DeclarationOriginMessageID.objects.get_or_create(
                    declaration_id=declaration.id,
                    defaults={'message_id': message_id})
                push_change_order_info_request(
                    self.application_type.orderId,
                    DECLARATION_RECEIVED_CODE,
                    DECLARATION_RECEIVED_COMMENT.format(
                        date=declaration.date.strftime(settings.DATE_FORMAT),
                        declaration_id=declaration.id),
                    message_id)
        except ValidationError as e:
            logger.error(f'{_Message.DATA_ERROR} ApplicationRequest '
                         f'({"; ".join(e.messages)})')
            return Response(
                order_id=self.application_type.orderId,
                org_code=_Message.values[_Message.DATA_ERROR],
                tech_code=CODE_ERROR,
                comment=_Message.DATA_ERROR)

        return Response(
            order_id=self.application_type.orderId,
            org_code=_Message.values[_Message.SUCCESS],
            tech_code=CODE_OK, comment=_Message.SUCCESS)

    def update_declaration(self, declaration):
        """Фиксирует изменения заявления.

        :return: Ответ
        :rtype: Response
        """

        change_map = Smev3ChangesMap()

        # Выполняет проверку изменений Желаемых ДОО
        unit_helper = Smev3DeclarationUnitChangeHelper(
            DeclarationUnit, 'DeclarationUnit')
        unit_helper.check_diff(self.application_type, declaration)

        # Выполняет проверку изменений Льгот
        benefits_helper = Smev3DeclarationPrivilegeChangeHelper(
            DeclarationPrivilege, 'DeclarationPrivilege')
        benefits_helper.check_diff(self.application_type, declaration)

        helpers = [
            unit_helper,
            benefits_helper,
        ]

        check_diff = [
            (Declaration, 'Declaration', declaration),
            (Children, 'Children', declaration.children),
        ]

        for children_delegate in declaration.children.childrendelegate_all:
            delegate = children_delegate.delegate
            check_diff.append((Delegate, 'Delegate', delegate))

        for model, model_name, instance in check_diff:
            helper = Smev3ChangeHelper(
                model, model_name, map_changes=change_map)
            helper.check_diff(self.application_type, instance)
            helpers.append(helper)

        if self.attachments:
            helper = Smev3DeclarationDocsChangeHelper(DocExtraInfo)
            helper.check_diff(declaration, self.attachments)
            helpers.append(helper)

        updated = (
            StorageHelper.create_change(
                declaration, helpers, source=ChangeSource.UPDATE_APPLICATION))

        if not updated:
            return Response(
                order_id=self.application_type.orderId,
                org_code=_Message.values[_Message.NO_CHANGES],
                tech_code=CODE_ERROR,
                comment=_Message.NO_CHANGES)
        else:
            changes = StorageHelper.get_change(updated)
            message_id = self.message.get('message_id')

            if message_id:
                push_change_order_info_request(
                    self.application_type.orderId,
                    CHANGE_DECLARATION_CODE,
                    CHANGE_DECLARATION_COMMENT.format(
                        changes=changes_to_str(changes)),
                    message_id)

        origin_id = OrderRequest.objects.filter(
            declaration=declaration,
            order_id__isnull=False
        ).values_list('order_id', flat=True).first()

        # Передается запрос OrderRequest с блоком UpdateOrderRequest
        # со статусом "Изменение заявление" в ЕПГУ
        if origin_id:
            UpdateOrderRequestSendingTask().apply_async(
                kwargs={
                    'declaration_id': declaration.id, 'order_id': origin_id,
                    'event': DECLARATION_CHANGED
                }
            )

        return Response(
            order_id=self.application_type.orderId,
            org_code=_Message.values[_Message.CHANGES_SUCCESS],
            tech_code=CODE_OK,
            comment=_Message.CHANGES_SUCCESS)

    def run(self):
        """Запускает процесс разбора полученных данных.

        :return: Ответ
        :rtype: Response
        """

        try:
            self.validate()
        except ValidationError as e:
            logger.error(f'{_Message.DATA_ERROR} ApplicationRequest '
                         f'({"; ".join(e.messages)})')
            return Response(
                order_id=self.application_type.orderId,
                org_code=_Message.ORG_CODE_150,
                tech_code=CODE_ERROR, comment=f'{e}')

        declaration = self.get_declaration_by_order_id()
        if declaration:
            return self.update_declaration(declaration)

        return self.create_declaration()


def update_middlename_params(middlename, params, is_parents=None):
    """
    Обновляет параметры устанавливая пол и тип представителя,
    если будут совпаданеия по окончанию отчества.
    Пол так же определяется и для ребенка.

    :param middlename: Отчество
    :type middlename: str
    :param params: Словарь с параметрами
    :type params: dict
    :param is_parents: Параметр, является ли родителем
    :type is_parents: bool or None
    """

    middlename = middlename.strip()

    if is_parents is not None:
        params['type'] = DelegateTypeEnumerate.LEX

    if middlename.endswith(('вна', 'кызы', 'ична')):
        params['gender'] = GenderEnumerate.FEMALE

        if is_parents:
            params['type'] = DelegateTypeEnumerate.MOTHER
    elif middlename.endswith(('ич', 'оглы')):
        params['gender'] = GenderEnumerate.MALE

        if is_parents:
            params['type'] = DelegateTypeEnumerate.FATHER

    return params


def changes_to_str(changes):
    """
    Возвращает изменения в виде строки в читаемом виде

    :param changes: Изменения
    :type changes: list
    :return: Строка с изменениями
    :rtype: str
    """

    return ' '.join(CHANGES_DATA_TEMPATE.format(index=index, **change)
                    for index, change in enumerate(changes, start=1))
