import datetime
import sys
import traceback

from celery import states
from celery.schedules import crontab
from django.db.models import OuterRef
from django.db.models import Subquery
from django.db.models import TextField
from django.db.models.functions import Cast
from django_celery_beat.models import PeriodicTask
from lxml import etree

from aio_client.base import RequestTypeEnum
from aio_client.base.exceptions import AioClientException
from aio_client.consumer.models import GetConsumerResponse
from aio_client.consumer.models import PostConsumerRequest
from aio_client.provider.api import get_requests
from educommon.ws_log.models import SmevLog
from educommon.ws_log.models import SmevSourceEnum

from concentrator.smev3_v321.constants import DECLARATION_AWAITING_CODE
from concentrator.smev3_v321.constants import DECLARATION_AWAITING_COMMENT
from concentrator.smev3_v321.constants import DECLARATION_REVIEWED_CODE
from concentrator.smev3_v321.constants import DECLARATION_REVIEWED_COMMENT
from concentrator.smev3_v321.helpers import push_change_order_info_request
from concentrator.smev3_v321.models import DeclarationOriginMessageID
from concentrator.smev3_v321.order.constants import CreateOrderStatusMapper
from kinder.core.async_tasks.tasks import PeriodicAsyncTask
from kinder.core.declaration.models import Declaration
from kinder.core.direct.models import DRS
from kinder.core.direct.models import DirectStatusLog

from concentrator import settings
from concentrator.smev3_v321.attachment_request.helpers import (
    AttachmentRequestHelper)
from concentrator.smev3_v321.base.utils import SMEV3RepositoryExecutors
from concentrator.smev3_v321.models import AttachmentRequest
from concentrator.smev3_v321.service_types import kinder_conc
from concentrator.smev3_v321 import settings as smev_settings


class Smev3PeriodicAsyncTask(PeriodicAsyncTask):
    """Периодическая таска, которая проверяет наличие в клиенте заявок
    с видом сведений message_type, запускает по ним формирование
    и отправку ответа.
    """

    message_type = settings.SMEV3_FORM_DATA_MESSAGE_TYPE
    run_every = smev_settings.SMEV3_FORM_DATA_TASK_RUN_EVERY

    description = 'Взаимодействие с формой-концентратором по СМЭВ 3'
    stop_executing = False
    LOG_TIME_FORMAT = '%d.%m.%Y %H:%M'

    _error_key = (
        'Ответ на сообщение {} не удалось отправить')

    def log(self, message, result=None, logging_data=None):
        """Логирование запроса СМЭВ 3.

        :param message: Сообщение
        :param result: Результат
        :param logging_data: Дополнительные параметры
        """

        logging_data = logging_data or {}

        data = {
            'service_address':
                RequestTypeEnum.get_url(RequestTypeEnum.PR_POST),
            'method_verbose_name': f'{self.message_type} ({self.description})',
            'direction': SmevLog.OUTGOING,
            'interaction_type': SmevLog.IS_SMEV,
            'source': SmevSourceEnum.CONCENTRATOR,
            'request': message.get('body'),
            'result': result,
            **logging_data
        }

        return SmevLog(**data).save()

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        values = {
            'Время начала':
                datetime.datetime.now().strftime(self.LOG_TIME_FORMAT),
        }
        self.set_progress(
            progress=f'Получаем запросы {self.message_type}',
            values=values)

        result = get_requests(self.message_type)
        values['Кол-во сообщений'] = str(len(result))
        error_list = []

        for message in result:
            origin_message_id = message['origin_message_id']
            values[f'Сообщение {origin_message_id}'] = (
                'Отправка ответа в СМЭВ')

            body = message.get('body')

            try:
                request_body = kinder_conc.parseString(body)
            except (
                etree.XMLSyntaxError, kinder_conc.GDSParseError, ValueError
            ) as exc:
                error_list.append(origin_message_id)
                error_message = (
                    f'Ошибка при разборе входящего сообщения - {exc}')
                values[self._error_key.format(origin_message_id)] = (
                    error_message)
                self.log(message, error_message)
                continue

            executor = SMEV3RepositoryExecutors.get_executor(request_body)
            if executor:
                try:
                    response = executor(message, request_body)
                    self.log(message, logging_data=response.logging_data)
                except AioClientException as exc:
                    error_list.append(origin_message_id)
                    error_message = (
                        f'{exc.message} (код ошибки - {exc.code})')
                    values[self._error_key.format(origin_message_id)] = (
                        error_message)
                    self.log(message, error_message)
                    continue
                except Exception:
                    error_list.append(origin_message_id)
                    error_message = (
                        '\n'.join(traceback.format_exception(*sys.exc_info())))
                    values[self._error_key.format(origin_message_id)] = (
                        error_message)
                    self.log(message, error_message)
                    continue
            else:
                error_list.append(origin_message_id)
                error_message = 'Неизвестный тип сообщения'
                values[self._error_key.format(origin_message_id)] = (
                    error_message)
                self.log(message, error_message)
                continue

            values[f'Сообщение {origin_message_id}'] = (
                'Отправка ответа в СМЭВ прошла успешно')

        if error_list:

            _key = 'Кол-во сообщений, по которым не удалось отправить ответ'
            values[_key] = str(len(error_list))

        values['Время окончания'] = (
            datetime.datetime.now().strftime(self.LOG_TIME_FORMAT))
        self.set_progress(
            progress='Завершено',
            task_state=states.SUCCESS,
            values=values)

        return self.state


class Smev3StatusChangePeriodicAsyncTask(PeriodicAsyncTask):
    """Периодическая таска, которая проверяет происходила ли смена статуса
    направлений на "Направлен в ДОО" с момента выполнения данной асинхронной
    задачи ранее и обрабатывает ответы AttachmentRequests
    """

    run_every = crontab(
        minute=settings.SMEV3_STATUS_CHANGE_TASK_EVERY_MINUTE,
        hour=settings.SMEV3_STATUS_CHANGE_TASK_EVERY_HOUR)

    service_name = AttachmentRequestHelper.service_name
    description = 'Проверка смены статуса направления'
    stop_executing = False
    LOG_TIME_FORMAT = '%d.%m.%Y %H:%M'
    SEND_REQUEST = 'Отправка запроса'
    GET_RESPONSE = 'Получение ответа по запросу'

    def log(self, message, message_type, result=None, logging_data=None):
        """Логирование запроса СМЭВ 3.

        :param message: Сообщение
        :param message_type: Тип сообщения
        :param result: Результат
        :param logging_data: Дополнительные параметры
        """

        logging_data = logging_data or {}

        data = {
            'service_address':
                RequestTypeEnum.get_url(RequestTypeEnum.PR_POST),
            'method_verbose_name': f'{message_type} {self.service_name}',
            'direction': SmevLog.OUTGOING,
            'interaction_type': SmevLog.IS_SMEV,
            'source': SmevSourceEnum.CONCENTRATOR,
            'request': message,
            'result': result,
            'method_name': self.service_name,
            **logging_data
        }

        return SmevLog(**data).save()

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        values = {
            'Время начала':
                datetime.datetime.now().strftime(self.LOG_TIME_FORMAT),
        }
        self.set_progress(
            progress='Проверка смены статусов направлений',
            values=values)

        last = PeriodicTask.objects.filter(name=self.name).last()
        directs = DirectStatusLog.objects.filter(
            status__code=DRS.REGISTER
        )
        if last.last_run_at:
            directs = directs.filter(created_at__gt=last.last_run_at)
        if not directs.exists():
            values.update({'Подходящих ДОО найдено': 0})
            self.set_progress(
                progress='Завершено',
                task_state=states.SUCCESS,
                values=values)

        units = directs.distinct('direct__group__unit').values_list(
            'direct__group__unit', flat=True)

        unit_counter = 0

        # отправка запросов для подходящих организаций
        for unit_id in units:
            helper = AttachmentRequestHelper(unit_id)
            request, attachments = helper.get_request()
            result = helper.send_request(request, attachments)
            status = 'Успешно' if result else 'Ошибка'
            unit_counter += 1
            self.log(request, self.SEND_REQUEST, result=status)

        # получение ответов с записью в СМЭВ лог
        requests_to_change = AttachmentRequest.objects.filter(
            response__isnull=True).annotate(
            message_id_str=Cast('message_id', output_field=TextField())
        ).annotate(resp=Subquery(GetConsumerResponse.objects.filter(
            origin_message_id=OuterRef('message_id_str')
        ).values('body')[:1])).annotate(
            req=Subquery(PostConsumerRequest.objects.filter(
                message_id=OuterRef('message_id_str')
            ).values('body')[:1])).filter(resp__isnull=False)

        for request in requests_to_change:
            request.response = request.resp
            request.save()
            self.log(request.req, self.GET_RESPONSE, result='Успешно',
                     logging_data={'response': request.resp})

        if unit_counter:
            values.update({'Подходящих ДОО найдено': unit_counter})
            self.set_progress(
                progress='Завершено',
                task_state=states.SUCCESS,
                values=values)

        return self.state


class DirectStatusCheckPeriodicAsyncTask(PeriodicAsyncTask):
    """
    Периодическая задача, которая отправляет информацию о статусе заявления
    """

    run_every = crontab(
        minute=settings.SMEV3_CHECK_DECLARATION_STATUS_CHANGES_MINUTE,
        hour=settings.SMEV3_CHECK_DECLARATION_STATUS_CHANGES_HOUR)

    description = (
        'Проверка необходимости отправки сообщений о статусе заявлений')

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)
        self.set_progress(
            progress='Началась отправка информации о статусах заявлений')
        sent = 0
        declarations_with_origin_message = Declaration.objects.filter(
            declarationoriginmessageid__isnull=False)

        not_reviewed_declarations = declarations_with_origin_message.filter(
            declarationoriginmessageid__reviewed_sent=False)

        for declaration in not_reviewed_declarations.iterator():
            if CreateOrderStatusMapper(
                declaration, None
            ).check_is_application_reviewed(declaration.status.code):
                message_id = declaration.declarationoriginmessageid.message_id
                comment = DECLARATION_REVIEWED_COMMENT.format(
                    **{'order_id': declaration.id,
                       'desired_date': declaration.desired_date.strftime(
                           settings.DATE_FORMAT)})
                push_change_order_info_request(
                    declaration.id, DECLARATION_REVIEWED_CODE,
                    comment, message_id)
                DeclarationOriginMessageID.objects.filter(
                    declaration=declaration).update(reviewed_sent=True)
                sent += 1

        awaiting_direction_declarations = (
            declarations_with_origin_message.filter(
                desired_date=datetime.date.today()))

        for declaration in awaiting_direction_declarations.iterator():
            if CreateOrderStatusMapper(
                declaration, None
            ).check_is_awaiting_direction(declaration.status.code):
                message_id = declaration.declarationoriginmessageid.message_id
                push_change_order_info_request(
                    declaration.id, DECLARATION_AWAITING_CODE,
                    DECLARATION_AWAITING_COMMENT, message_id)
                sent += 1

        self.set_progress(
            progress='Завершено',
            task_state=states.SUCCESS,
            values={'Отправлено сообщений': sent})

        return self.state
