from abc import ABC
from abc import abstractmethod
from collections import Iterable
from io import StringIO

from kinder.core.dict.models import GroupOrientationDocuments
from kinder.core.dict.models import GroupSpec
from kinder.core.declaration_status.enum import DSS
from kinder.core.unit.models import Unit
from kinder.core.unit.models import UnitKind

from concentrator import settings

from .rules import DULDelegateTypeRule
from .rules import HealthNeedRule
from .rules import GroupTypeRule
from .rules import WorkTypeRule


class SMEV3Response:

    def __init__(self, response, logging_data):
        """Настройка ответа.

        :param response: ответ
        :type response: requests.Response
        :param logging_data: дополнительные данные для логирования
        :type logging_data: dict
        """

        self.response = response
        self.logging_data = logging_data


class BaseExecutor(ABC):
    """Базовый класс исполнителя сервиса СМЭВ 3."""

    name_service = None
    type_service = None

    @classmethod
    def __subclasshook__(cls, subclass):

        if cls is BaseExecutor:
            return any(
                True for sub in subclass.__mro__
                if sub.__dict__.get('process') is not None)

        return NotImplemented

    @classmethod
    @abstractmethod
    def process(cls, message, request_body):
        """Запускает выполнение логики обработки сервиса.

        :param message: xml пакет
        :param request_body: тип

        :return: инстанс ответа
        :rtype: SMEV3Response
        """
        pass


class SMEV3RepositoryExecutors:
    """Хранилище исполнителей сервисов СМЭВ 3."""

    _store = [
        # (Наименование сервиса, Тип сервиса, Обработчик сервиса),
        # ...
    ]

    @classmethod
    def set_up(cls, config):
        """Инициализирует хранилище класса.

        Принимает список наследников класса BaseExecutor
        или структуру вида:
        (Наименование сервиса, Тип сервиса, Обработчик сервиса).

        :param config: список настроек
        :type config: Union[
            Iterable[Tuple[str, Type, callable]],
            Iterable[BaseExecutor]]
        :return: None
        """
        cls._store.extend(
            (i.name_service, i.type_service, i.process)
            if not isinstance(i, Iterable) and issubclass(i, BaseExecutor)
            else i for i in config)

    @classmethod
    def get_executor(cls, message):
        """Определяет исполнителя для полученного пакета.

        :param message: Тип, содержащий данные сервиса
        :return: Ссылку на метод исполнителя
        """

        for name_service, type_service, executor in cls._store:
            service = getattr(message, name_service, None)
            if (service
                    and isinstance(service, type_service)
                    and callable(executor)):
                return executor


def render_type2xml(type_, name_type, pretty_print=False):
    """Рендерит классы в xml.

    :param type_: инстанс типа
    :param name_type: наименование типа
    :param pretty_print: признак "красивого" вывода в xml
    :return: сформированный xml
    """

    buffer = StringIO()
    type_.export(
        buffer, 0, name_=name_type, pretty_print=pretty_print)

    return buffer.getvalue()


def is_cancel_allowed(status):
    """Проверяет разрешена ли отмена заявления в указанном статусе

    :param status: Статус заявления
    :type status: DeclarationStatus
    :return: Разрешена ли отмена
    :rtype: bool
    """

    if status.code in list(DSS.values.keys()):
        if getattr(settings, status.code.upper(), False):
            return True
    elif settings.OTHER_STATUS:
        return True

    return False


def get_oktmo_region():
    """Возвращает ОКТМО головного учреждения.

    :return: октмо или None
    :rtype: Optional[str]
    """

    region = Unit.objects.filter(kind=UnitKind.REGION).last()

    return region.octmo if region else None


def get_identity_doc_name(dul_type):
    """Возвращает данные для блока IdentityDocName.

    :param dul_type: Тип документа представителя
    :type dul_type: DULDelegateType
    :return: словарь с данными
    :rtype: dict
    """

    if dul_type:
        identity_doc_name_params = {
            'code': (
                dul_type.esnsi_code
                or DULDelegateTypeRule.get_service_list(int(dul_type.code))),
            'valueOf_': dul_type.name
        }
    else:
        identity_doc_name_params = {'code': None, 'valueOf_': None}

    return identity_doc_name_params


def get_adaptation_group_type(health_need, work_type_code):
    """Возвращает данные для блока AdaptationGroupType.

    :param health_need: Специфика группы
    :type health_need: HealthNeed
    :param work_type_code: Код направленности группы
    :type work_type_code: str
    :return: словарь с данными
    :rtype: dict
    """

    adaptation_group_type_params = {}
    if health_need:
        adaptation_group_type_params = {
            'code': (
                    health_need.esnsi_code
                    or HealthNeedRule.get_service(
                        (health_need.code, work_type_code))),
            'valueOf_': health_need.name
        }

    return adaptation_group_type_params


def get_adaptation_group(desired_group_type):
    """Возвращает данные для блока AdaptationGroup.

    :param desired_group_type: Направленность группы
    :type desired_group_type: GroupType
    :return: словарь с данными
    :rtype: dict
    """

    if desired_group_type:

        adaptation_group_params = {
            'code': (
                    desired_group_type.esnsi_code
                    or GroupTypeRule.get_service_list(desired_group_type.code)
            ),
            'valueOf_': desired_group_type.name
        }
    else:
        adaptation_group_params = {'code': None, 'valueOf_': None}

    return adaptation_group_params


def get_schedule(work_type):
    """Возвращает данные для блока Schedule.

    :param work_type: Режим работы
    :type work_type: WorkType
    :return: словарь с данными
    :rtype: dict
    """

    if work_type:
        schedule_params = {
            'code': (
                    work_type.esnsi_code
                    or WorkTypeRule.get_service(work_type.code)),
            'valueOf_': work_type.name
        }
    else:
        schedule_params = {'code': None, 'valueOf_': None}

    return schedule_params


def get_language(spec):
    """Возвращает данные для блока Language.

    :param spec: Язык обучения
    :type spec: GroupSpec
    :return: словарь с данными
    :rtype: dict
    """

    if spec:
        language_params = {
            'code': spec.esnsi_code,
            'valueOf_': spec.name
        }
    else:
        language_code, language_value = (
                GroupSpec.objects.filter(
                    code=GroupSpec.RUS
                ).values_list('esnsi_code', 'name').first() or (None, None))
        language_params = {
            'code': language_code,
            'valueOf_': language_value
        }

    return language_params


def get_medical_report(child):
    """Возвращает данные для блока DocName (MedicalReport).

    :param child: Ребенок
    :type child: Children
    :return: словарь с данными
    :rtype: dict
    """

    medical_report_params = {'code': None, 'valueOf_': None}

    if child.health_need_confirmation:
        group_orientation_document = (
            GroupOrientationDocuments.objects.filter(
                id=child.health_need_confirmation).first())
        if group_orientation_document:
            medical_report_params.update({
                'code': group_orientation_document.esnsi_code,
                'valueOf_': group_orientation_document.name
            })

    return medical_report_params
