from django.db.models import OuterRef
from django.db.models import Prefetch
from django.db.models import Subquery

from aio_client.base import const as aio_base_const
from aio_client.provider.api import PostProviderRequest
from aio_client.provider.api import push_request
from kinder.core.children.models import ChildrenDelegate
from kinder.core.declaration.models import Declaration
from kinder.core.direct.models import Direct
from kinder.core.direct.models import DRS
from kinder.core.utils.fias import name_from_guid
from kinder.webservice.api.declaration import get_declaration_by_client_id
from kinder.webservice.api.exceptions import ApiException

from concentrator.smev3_v321.base import constants as base_enums
from concentrator.smev3_v321.base.exceptions import ContentFailure
from concentrator.smev3_v321.base.utils import BaseExecutor
from concentrator.smev3_v321.base.utils import SMEV3Response
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.base.utils import get_identity_doc_name
from concentrator.smev3_v321.base.utils import get_adaptation_group_type
from concentrator.smev3_v321.base.utils import get_adaptation_group
from concentrator.smev3_v321.base.utils import get_schedule
from concentrator.smev3_v321.base.utils import get_language
from concentrator.smev3_v321.base.utils import get_medical_report
from concentrator.smev3_v321.service_types import kinder_conc


class GetApplicationAdmissionRequestExecutor(BaseExecutor):
    """Исполнитель сервиса GetApplicationAdmissionRequest."""

    name_service = 'GetApplicationAdmissionRequest'
    type_service = kinder_conc.GetApplicationAdmissionRequestType

    @classmethod
    def prepare_query(cls):
        """Выполняет подготовку запроса."""

        subquery_first_delegate = (
            ChildrenDelegate.objects.filter(
                id=Subquery(
                    ChildrenDelegate.objects.filter(
                        children_id=OuterRef('children_id')
                    ).order_by('id').values('id')[:1])
            ).select_related('delegate'))

        subquery_unit_code = (
            Subquery(Direct.objects.filter(
                declaration_id=OuterRef('id'),
                status__code__in=(DRS.REGISTER, DRS.DOGOVOR),
                group__unit__is_not_show_on_poral=False
            ).values('group__unit__code')[:1]))

        query = (
            Declaration.objects.annotate(
                unit_code=subquery_unit_code
            ).select_related(
                'children'
            ).filter(
                unit_code__isnull=False
            ).prefetch_related(
                Prefetch(
                    'children__childrendelegate_set',
                    queryset=subquery_first_delegate,
                    to_attr='first_childrendelegate')))

        return query

    @classmethod
    def process(cls, message, request_body):

        try:
            declaration = get_declaration_by_client_id(
                cls.prepare_query(),
                request_body.GetApplicationAdmissionRequest.orderId)
        except ApiException:
            declaration = None

        try:
            response_body = render_type2xml(
                cls.get_response(declaration), name_type='FormDataResponse')
            content_failure_code = None
            content_failure_comment = None
        except ContentFailure as exc:
            response_body = 'null'
            content_failure_code = exc.content_failure_code
            content_failure_comment = exc.content_failure_comment

        response = push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            attachments=None,
            content_failure_code=content_failure_code,
            content_failure_comment=content_failure_comment or '',
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message']
        ))

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': content_failure_comment or base_enums.SUCCESS_MESSAGE
            })

    @classmethod
    def get_response(cls, declaration):
        """Формирует ответ на GetApplicationAdmissionRequest.

        :param declaration: Заявление
        :type declaration: Declaration
        :return: ответ
        :rtype: FormDataResponseType
        :raise: ContentFailure
        """

        if not declaration:
            raise ContentFailure(
                aio_base_const.FAILURE,
                'Заявление по указанным параметрам не найдено')

        child = declaration.children
        children_delegate, *_ = child.first_childrendelegate or (None,)
        if not children_delegate:
            raise ContentFailure(
                aio_base_const.FAILURE, 'Не найден представитель')

        delegate = children_delegate.delegate

        person_info = kinder_conc.PersonInfoType(
            PersonSurname=delegate.surname,
            PersonName=delegate.firstname,
            PersonMiddleName=delegate.patronymic,
            PersonPhone=delegate.phones,
            PersonEmail=delegate.email)

        person_identity_doc_info = kinder_conc.PersonIdentityDocInfoType(
            IdentityDocName=kinder_conc.DataElementType(
                **get_identity_doc_name(delegate.dul_type)),
            IdentityDocSeries=delegate.dul_series,
            IdentityDocNumber=delegate.dul_number,
            IdentityDocIssueDate=delegate.dul_date,
            IdentityDocIssueCode=delegate.dul_place,
            IdentityDocIssued=delegate.dul_issued_by)

        child_info = kinder_conc.ChildInfoType(
            ChildSurname=child.surname,
            ChildName=child.firstname,
            ChildMiddleName=child.patronymic,
            ChildBirthDate=child.date_of_birth,
            ChildBirthDocRF=kinder_conc.ChildBirthDocRFType(
                ChildBirthDocSeries=child.dul_series,
                ChildBirthDocNumber=child.dul_number,
                ChildBirthDocIssueDate=child.dul_date,
                ChildBirthDocActNumber=child.zags_act_number,
                ChildBirthDocActDate=child.zags_act_date,
                ChildBirthDocIssued=child.zags_act_place))

        address_info = kinder_conc.AddressType(
            FullAddress=child.reg_address_full,
            Region=kinder_conc.DataElementType(
                code=child.reg_address_place,
                valueOf_=name_from_guid(child.reg_address_place)),
            Street=kinder_conc.DataElementType(
                code=child.reg_address_street,
                valueOf_=name_from_guid(child.reg_address_street)),
            House=kinder_conc.DataElementType(
                code=child.reg_address_house_guid,
                valueOf_=child.reg_address_house),
            Building1=child.reg_address_corps,
            Apartment=child.reg_address_flat)

        entry_params = kinder_conc.EntryParamsType(
            EntryDate=declaration.desired_date,
            Language=kinder_conc.DataElementType(
                **get_language(declaration.spec)),
            Schedule=kinder_conc.DataElementType(
                **get_schedule(declaration.work_type)),
            AgreementOnFullDayGroup=declaration.consent_full_time_group)

        adaptation_program_params = {}
        if child.health_need:

            adaptation_program_params['AdaptationGroupType'] = (
                kinder_conc.DataElementType(
                    **get_adaptation_group_type(
                        child.health_need, declaration.work_type.code)))

        adaptation_program = kinder_conc.AdaptationProgramType(
            AdaptationGroup=kinder_conc.DataElementType(
                **get_adaptation_group(declaration.desired_group_type)),
            AgreementOnGeneralGroup=declaration.consent_dev_group,
            AgreementOnCareGroup=declaration.consent_care_group,
            NeedSpecialCareConditions=child.health_need_special_support,
            **adaptation_program_params)

        medical_report_params = {}

        if child.health_series:
            medical_report_params['DocSeries'] = child.health_series

        if child.health_need_expiration_date:
            medical_report_params['DocExpirationDate'] = (
                child.health_need_expiration_date)

        medical_report = kinder_conc.MedicalReportWithoutFilesType(
            DocName=kinder_conc.DataElementType(**get_medical_report(child)),
            DocNumber=child.health_number,
            DocIssueDate=child.health_need_start_date,
            DocIssued=child.health_issued_by, **medical_report_params)

        get_application_admission_response = (
            kinder_conc.GetApplicationAdmissionResponseType(
                orderId=int(declaration.client_id),
                PersonInfo=person_info,
                PersonIdentityDocInfo=person_identity_doc_info,
                ChildInfo=child_info,
                Address=address_info,
                EntryParams=entry_params,
                AdaptationProgram=adaptation_program,
                MedicalReport=medical_report,
                EduOrganizationCode=declaration.unit_code))

        response = kinder_conc.FormDataResponseType(
            GetApplicationAdmissionResponse=get_application_admission_response)

        return response
