from aio_client.provider.api import push_request
from aio_client.provider.models import PostProviderRequest
from django.db.models import OuterRef
from django.db.models import Prefetch
from django.db.models import Subquery

from concentrator.smev3_v321.base import constants
from concentrator.smev3_v321.base.exceptions import ContentFailure
from concentrator.smev3_v321.base.utils import BaseExecutor
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.base.utils import SMEV3Response
from concentrator.smev3_v321.constants import ERROR_ORG_CODE
from concentrator.smev3_v321.service_types import kinder_conc
from kinder.core.children.helpers import get_zipcode
from kinder.core.children.models import ChildrenDelegate
from kinder.core.children.models import DelegateTypeEnumerate
from kinder.core.declaration.enum import DeclarationSourceEnum
from kinder.core.declaration.enum import DeclarationTypeInteractionEnum
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationPrivilege
from kinder.core.declaration.models import DeclarationUnit
from kinder.core.dict.models import DULTypeEnumerate
from kinder.core.utils.fias import name_from_guid
from kinder.webservice.api.declaration import get_declaration_by_client_id
from kinder.webservice.api.exceptions import ApiException

from concentrator.smev3_v321.base.utils import get_identity_doc_name
from concentrator.smev3_v321.base.utils import get_schedule
from concentrator.smev3_v321.base.utils import get_language
from concentrator.smev3_v321.base.utils import get_adaptation_group_type
from concentrator.smev3_v321.base.utils import get_adaptation_group
from concentrator.smev3_v321.base.utils import get_medical_report


class GetApplicationRequestExecutor(BaseExecutor):
    """Исполнитель сервиса GetApplicationRequest."""

    name_service = 'GetApplicationRequest'
    type_service = kinder_conc.GetApplicationRequestType

    @classmethod
    def prepare_query(cls):
        """Выполняет подготовку запроса."""

        subquery_first_delegate = (
            ChildrenDelegate.objects.filter(
                id=Subquery(
                    ChildrenDelegate.objects.filter(
                        children_id=OuterRef('children_id')
                    ).order_by('id').values('id')[:1])
            ).select_related('delegate'))

        subquery_declaration_units = (
            DeclarationUnit.objects.select_related('unit', 'sibling').only(
                'unit__name', 'unit_id', 'ord', 'declaration_id', 'sibling'))

        query = (
            Declaration.objects.select_related(
                'children'
            ).prefetch_related(
                Prefetch('declarationunit_set',
                         queryset=subquery_declaration_units,
                         to_attr='declaration_units')
            ).prefetch_related(
                Prefetch('children__childrendelegate_set',
                         queryset=subquery_first_delegate,
                         to_attr='first_childrendelegate')
            ).filter(
                source=DeclarationSourceEnum.CONCENTRATOR,
                type_interaction=DeclarationTypeInteractionEnum.SMEV_3,
            )
        )

        return query

    @classmethod
    def process(cls, message, request_body):

        try:
            declaration = get_declaration_by_client_id(
                cls.prepare_query(),
                request_body.GetApplicationRequest.orderId)
        except ApiException:
            declaration = None

        try:
            response_body = render_type2xml(
                cls.get_response(declaration), name_type='FormDataResponse')
            content_failure_code = None
            content_failure_comment = None
        except ContentFailure as exc:
            response_body = render_type2xml(
                kinder_conc.FormDataResponseType(
                    changeOrderInfo=kinder_conc.changeOrderInfoType(
                        orderId=kinder_conc.orderIdType(
                            request_body.GetApplicationRequest.orderId),
                        statusCode=kinder_conc.statusCodeType(
                            exc.content_failure_code),
                        comment=exc.content_failure_comment)),
                name_type='FormDataResponse')
            content_failure_code = exc.content_failure_code
            content_failure_comment = exc.content_failure_comment

        response = push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            attachments=None,
            content_failure_code=content_failure_code,
            content_failure_comment=content_failure_comment or '',
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message']
        ))

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': content_failure_comment or constants.SUCCESS_MESSAGE
            })

    @classmethod
    def get_response(cls, declaration):
        """Формирует ответ на GetApplicationRequest.

        :param declaration: Заявление
        :type declaration: Declaration
        :return: Ответ
        :rtype: FormDataResponseType
        :raise: ContentFailure
        """

        if not declaration:
            raise ContentFailure(
                ERROR_ORG_CODE, 'Заявление по указанным параметрам не найдено')

        child = declaration.children
        children_delegate, *_ = (
            declaration.children.first_childrendelegate or (None,))

        if not children_delegate:
            raise ContentFailure(
                ERROR_ORG_CODE, 'Не найден представитель')

        delegate = children_delegate.delegate

        person_identity_doc = kinder_conc.PersonIdentityDocInfoType(
            IdentityDocName=kinder_conc.DataElementType(
                **get_identity_doc_name(delegate.dul_type)),
            IdentityDocSeries=delegate.dul_series,
            IdentityDocNumber=delegate.dul_number,
            IdentityDocIssueDate=delegate.dul_date,
            IdentityDocIssueCode=delegate.dul_place,
            IdentityDocIssued=delegate.dul_issued_by,
        )
        parents = delegate.type in DelegateTypeEnumerate.PARENTS
        person_info = kinder_conc.PersonInfoType(
            PersonSurname=delegate.surname,
            PersonName=delegate.firstname,
            PersonMiddleName=delegate.patronymic,
            PersonPhone=delegate.phones,
            PersonEmail=delegate.email,
            Parents=parents,
        )

        if not parents:
            other_representative = kinder_conc.OtherRepresentativeType(
                OtherRepresentativeDocName=getattr(
                    delegate.dul_type, 'name', None),
                OtherRepresentativeDocSeries=delegate.dul_series,
                OtherRepresentativeDocNumber=delegate.dul_number,
                OtherRepresentativeDocDate=delegate.dul_date,
                OtherRepresentativeDocIssued=delegate.dul_issued_by,
            )
            person_info.set_OtherRepresentative(other_representative)

        child_info = kinder_conc.ChildInfoType(
            ChildSurname=child.surname,
            ChildName=child.firstname,
            ChildMiddleName=child.patronymic,
            ChildBirthDate=child.date_of_birth,
            ChildSNILS=child.snils,
        )

        if child.dul_type == DULTypeEnumerate.SVID:
            child_doc_rf = kinder_conc.ChildBirthDocRFType(
                ChildBirthDocSeries=child.dul_series,
                ChildBirthDocNumber=child.dul_number,
                ChildBirthDocIssueDate=child.dul_date,
                ChildBirthDocActNumber=child.zags_act_number,
                ChildBirthDocActDate=child.zags_act_date,
                ChildBirthDocIssued=child.zags_act_place,
            )
            child_info.set_ChildBirthDocRF(child_doc_rf)
        else:
            child_doc_foreign = kinder_conc.ChildBirthDocForeignType(
                ChildBirthDocName=child.dul_type.name,
                ChildBirthDocSeries=child.dul_series,
                ChildBirthDocNumber=child.dul_number,
                ChildBirthDocIssueDate=child.dul_date,
                ChildBirthDocIssued=child.zags_act_place,
            )
            child_info.set_ChildBirthDocForeign(child_doc_foreign)

        address_info = kinder_conc.AddressType(
            FullAddress=child.reg_address_full,
            Index=get_zipcode(child.reg_address_full),
            City=kinder_conc.DataElementType(
                code=child.reg_address_place,
                valueOf_=name_from_guid(child.reg_address_place)),
            Street=kinder_conc.DataElementType(
                code=child.reg_address_street,
                valueOf_=name_from_guid(child.reg_address_street)),
            House=kinder_conc.DataElementType(
                code=child.reg_address_house_guid,
                valueOf_=child.reg_address_house),
            Building1=child.reg_address_corps,
            Apartment=child.reg_address_flat,
        )

        entry_params = kinder_conc.EntryParamsType(
            EntryDate=declaration.desired_date,
            Language=kinder_conc.DataElementType(
                **get_language(declaration.spec)),
            Schedule=kinder_conc.DataElementType(
                **get_schedule(declaration.work_type)),
            AgreementOnFullDayGroup=declaration.consent_full_time_group,
        )

        edu_org_info = kinder_conc.EduOrganizationsType(
            AllowOfferOther=declaration.offer_other)
        all_declaration_units = declaration.declaration_units

        for declaration_unit in all_declaration_units:
            edu_org_info.add_EduOrganization(
                kinder_conc.EduOrganizationType(
                    code=declaration_unit.unit_id,
                    PriorityNumber=declaration_unit.ord,
                    valueOf_=declaration_unit.unit.name,
                )
            )

        adaptation_program_params = {}

        if declaration.desired_group_type:

            adaptation_program_params['AdaptationGroup'] = (
                kinder_conc.DataElementType(
                    **get_adaptation_group(declaration.desired_group_type)))

        if child.health_need:
            adaptation_program_params['AdaptationGroupType'] = (
                kinder_conc.DataElementType(
                    **get_adaptation_group_type(
                        child.health_need, declaration.work_type.code)))

        adaptation_program = kinder_conc.AdaptationProgramType(
            AgreementOnGeneralGroup=declaration.consent_dev_group,
            AgreementOnCareGroup=declaration.consent_care_group,
            NeedSpecialCareConditions=child.health_need_special_support,
            **adaptation_program_params,
        )

        medical_report = kinder_conc.MedicalReportWithoutFilesType(
            DocName=kinder_conc.DataElementType(**get_medical_report(child)),
            DocSeries=child.health_series,
            DocNumber=child.health_number,
            DocIssueDate=child.health_need_start_date,
            DocIssued=child.health_issued_by,
            DocExpirationDate=child.health_need_expiration_date,
        )

        app_type = kinder_conc.GetApplicationResponseType(
            orderId=int(declaration.client_id),
            PersonInfo=person_info,
            PersonIdentityDocInfo=person_identity_doc,
            ChildInfo=child_info,
            Address=address_info,
            EntryParams=entry_params,
            AdaptationProgram=adaptation_program,
            MedicalReport=medical_report,
            EduOrganizations=edu_org_info,
        )

        declaration_privilege = DeclarationPrivilege.objects.filter(
            privilege=declaration.best_privilege, declaration=declaration
        ).first()

        if declaration_privilege:
            doc_info = {}

            if declaration_privilege.doc_type:
                doc_info['DocName'] = kinder_conc.DataElementType(
                    code=declaration_privilege.privilege.esnsi_code,
                    valueOf_=declaration_privilege.doc_type.name
                )

            if declaration_privilege.doc_series:
                doc_info['DocSeries'] = declaration_privilege.doc_series

            if declaration_privilege.doc_number:
                doc_info['DocNumber'] = declaration_privilege.doc_number

            doc_info_content = {}

            if doc_info:
                doc_info_content['BenefitDocInfo'] = kinder_conc.DocInfoType(
                    **doc_info
                )

            app_type.set_BenefitInfo(
                kinder_conc.BenefitInfoWithoutFilesType(
                    BenefitCategory=kinder_conc.DataElementType(
                        code=declaration_privilege.privilege.esnsi_code,
                        valueOf_=declaration_privilege.privilege.name,
                    ),
                    **doc_info_content
                )
            )

        for declaration_unit in all_declaration_units:
            if declaration_unit.sibling:
                app_type.add_BrotherSisterInfo(
                    kinder_conc.BrotherSisterInfoType(
                        ChildSurname=declaration_unit.sibling.surname,
                        ChildName=declaration_unit.sibling.firstname,
                        ChildMiddleName=declaration_unit.sibling.patronymic,
                        EduOrganization=kinder_conc.DataElementType(
                            code=declaration_unit.unit_id,
                            valueOf_=declaration_unit.unit.name,
                        ),
                    )
                )

        response = kinder_conc.FormDataResponseType(
            GetApplicationResponse=app_type)

        return response
