import datetime

from aio_client.provider.api import push_request
from aio_client.provider.models import PostProviderRequest
from django.conf import settings

from concentrator.settings import SMEV3_FORM_DATA_MESSAGE_TYPE
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.constants import DECLARATION_STATUS_MAP
from concentrator.smev3_v321.constants import DEFAULT_CODE_COMMENT
from concentrator.smev3_v321.constants import DIRECT_STATUS_MAP
from concentrator.smev3_v321.constants import DIRECTED_WITHOUT_RESPONSE
from concentrator.smev3_v321.models import ApplicantAnswer
from concentrator.smev3_v321.service_types import kinder_conc
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationStatusLog
from kinder.core.declaration_status.models import DeclarationStatusTransfer
from kinder.core.direct.models import DirectRefusalReason
from kinder.core.direct.models import DirectStatusLog
from kinder.core.direct.models import DRS
from kinder.core.direct.models import RefusalReasonInitiatorType
from kinder.core.group.models import Pupil


def get_code_and_comment(instance):
    """
    Возвращает код и комментарий для передачи сообщения при смене статуса
    заявления или направления

    :param instance: Заявление или направление
    :type instance: Union[Declaration, Direct]
    :return: Код и комментарий
    :rtype: Tuple[int, str]
    """
    
    if isinstance(instance, Declaration):
        code, comment = DECLARATION_STATUS_MAP.get(
            instance.status.code, DEFAULT_CODE_COMMENT)
        decl_status_log = DeclarationStatusLog.objects.filter(
            declaration=instance, status=instance.status
        ).latest('datetime')
        status_transfer = DeclarationStatusTransfer.objects.filter(
            from_status=decl_status_log.old_status,
            to_status=decl_status_log.status
        ).first()
        if status_transfer and status_transfer.auto_change_days:
            date_before = datetime.date.today() + datetime.timedelta(
                days=status_transfer.auto_change_days)
            doc_date_before = (
                f' в срок до {date_before.strftime(settings.DATE_FORMAT)} ')
        else:
            doc_date_before = ''

        data = {
            'mo': instance.mo.name,
            'comment': decl_status_log.comment,
            'mo_address': instance.mo.address_full,
            'doc_date_before': doc_date_before,
        }
    else:
        status_log = DirectStatusLog.objects.filter(
            direct=instance, status=instance.status
        ).latest('created_at')
        if instance.group.unit.days_for_direct_dogovor:
            dogovor_days = status_log.created_at + datetime.timedelta(
                days=instance.group.unit.days_for_direct_dogovor)
        else:
            dogovor_days = None
        data = {
            'doo': instance.group.unit.name,
            'group': instance.group.name,
            'age_cat': instance.group.age_cat.name,
            'document_details': instance.document_details,
            'mo': instance.declaration.mo.name,
            'unit_address': instance.group.unit.address_full,
        }
        if instance.status.code == DRS.REFUSED:
            reason = DirectRefusalReason.objects.filter(
                direct_status_log=status_log,
                reason__initiator_type=RefusalReasonInitiatorType.PARENT
            ).first()
            if (reason and reason.reason.initiator_type ==
                    RefusalReasonInitiatorType.EMPLOYEE):
                data.update({'reason': reason.reason.name})
                code, comment = DIRECT_STATUS_MAP.get(
                    DRS.REJECT, DEFAULT_CODE_COMMENT)
            else:
                code, comment = DIRECT_STATUS_MAP.get(
                    DRS.NEED_CHANGE, DEFAULT_CODE_COMMENT)
        elif instance.status.code == DRS.REGISTER:
            if ApplicantAnswer.objects.filter(direct=instance).exists():
                code, comment = DIRECT_STATUS_MAP.get(instance.status.code)
            else:
                code, comment = DIRECT_STATUS_MAP.get(
                    DIRECTED_WITHOUT_RESPONSE)
        else:
            if status_log.old_status.code == DRS.REGISTER:
                date_before = status_log.created_at + datetime.timedelta(
                    days=instance.group.unit.days_for_reject_direct)
                data.update(
                    {'reason':
                     ' окончания срока действия направления '
                     f'{date_before.strftime(settings.DATE_FORMAT)}'})
            elif status_log.old_status.code == DRS.DOGOVOR:
                if dogovor_days is not None:
                    reason = (' окончания срока действия направления '
                              f'{dogovor_days.strftime(settings.DATE_FORMAT)}')
                else:
                    reason = ''
                data.update({'reason': reason})
            if dogovor_days is not None:
                data.update(
                    {'dogovor_days':
                     f' до {dogovor_days.strftime(settings.DATE_FORMAT)}'})
            code, comment = DIRECT_STATUS_MAP.get(
                instance.status.code, DEFAULT_CODE_COMMENT)

        pupil = Pupil.objects.filter(
            grup=instance.group, children=instance.declaration.children
        ).first()

        if pupil:
            data.update({'date_in_order_to_doo': pupil.date_in_order_to_doo})

        if instance.group.unit.days_for_reject_direct:
            date_before = status_log.created_at + datetime.timedelta(
                days=instance.group.unit.days_for_reject_direct)
            data.update(
                {'date_before':
                 f' до {date_before.strftime(settings.DATE_FORMAT)}'})
    return code, comment.format(**data)


def push_change_order_info_request(order_id, org_code, comment, message_id):
    """Отправляет сообщение ChangeOrderInfo"""

    response_body = render_type2xml(
        kinder_conc.FormDataResponseType(
            changeOrderInfo=kinder_conc.changeOrderInfoType(
                orderId=kinder_conc.orderIdType(order_id),
                statusCode=kinder_conc.statusCodeType(
                    orgCode=org_code),
                comment=comment)),
        name_type='FormDataResponse')

    return push_request(PostProviderRequest(
        origin_message_id=message_id,
        body=response_body,
        message_type=SMEV3_FORM_DATA_MESSAGE_TYPE,
        attachments=None,
        content_failure_code=None,
        content_failure_comment='',
        replay_to=message_id,
        is_test_message=False))
