from kinder.core.children.models import Children
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationUnit
from kinder.core.dict.models import UnitKind
from kinder.core.direct.models import DirectStatusLog
from kinder.core.observer import ChangesObserver
from kinder.core.unit.models import Unit
from kinder.core.unit.models import UnitStatus
from kinder.core.dict.models import UnitKind
from .constants import DECLARATION_CHANGES_REFUSED_CODE
from .constants import DECLARATION_CHANGES_REFUSED_COMMENT
from .helpers import get_code_and_comment
from .helpers import push_change_order_info_request
from ..models import ChangeDeclaration
from ..models import ChangeStatus
from ..settings import SMEV3_FORM_DATA_MESSAGE_TYPE
from .application_order_info_request.constants import (
    APPLICATION_ORDER_INFO_REQUEST_SUCCESS_CODE)
from .application_order_info_request.constants import (
    APPLICATION_ORDER_INFO_REQUEST_SUCCESS_COMMENT)
from .application_order_info_request.helpers import (
    ApplicationOrderInfoRequestExecutor)
from .esnsi.classifiers import MaxDooClassifier
from .esnsi.classifiers import UnitClassifier
from .esnsi.helpers import ESNSIUnitTaskFactory
from .esnsi.helpers import get_classifier_instance
from .models import DeclarationOriginMessageID
from .models import OrderRequest
from .models import UpdateClassifierRequest
from .order.tasks import UpdateOrderRequestSendingTask


class DeclarationStatusChangesObserver(ChangesObserver):
    """
    Класс для отправки запроса ApplicationOrderInfoRequest при изменении
    статуса заявления.
    """

    model = Declaration
    tracked_fields = ['status']

    def post_save(self, instance, created, **kwargs):
        """
        Отправляет запрос ApplicationOrderInfoRequest при изменении
        статуса заявления
        """

        origin_id = OrderRequest.objects.filter(
            declaration=instance,
            order_id__isnull=False
        ).values_list('order_id', flat=True).first()

        if self.has_changes(instance) and origin_id and not created:
            UpdateOrderRequestSendingTask().apply_async(
                kwargs={
                    'declaration_id': instance.id,
                    'order_id': origin_id,
                    'decl_status_changed': True
                }
            )

        declaration_message_id = DeclarationOriginMessageID.objects.filter(
            declaration=instance).first()
        if created or not declaration_message_id:
            return

        message_id = declaration_message_id.message_id

        org_code, comment = get_code_and_comment(instance)
        push_change_order_info_request(
            instance.id, org_code, comment, message_id)


class ChangeDeclarationObserver(ChangesObserver):
    """
    Класс для отправки запроса при отклонении изменений заявления
    """

    model = ChangeDeclaration
    tracked_fields = ['state']

    def post_save(self, instance, created, **kwargs):
        """
        Отправляет запрос ApplicationOrderInfoRequest при отклонении
        изменений заявления
        """

        declaration_message_id = DeclarationOriginMessageID.objects.filter(
            declaration=instance.declaration).first()

        if (created or not declaration_message_id and
                instance.state != ChangeStatus.REJECT):
            return

        message_id = declaration_message_id.message_id
        org_code = DECLARATION_CHANGES_REFUSED_CODE
        comment = DECLARATION_CHANGES_REFUSED_COMMENT.format(
            comment=instance.commentary)
        push_change_order_info_request(
            instance.declaration.id, org_code, comment, message_id)


class DeclarationFieldsChangesObserver(ChangesObserver):
    """
    Класс для отправки запроса UpdateOrderRequest при изменении полей
    заявления, кроме статуса, используемых в сервисе OrderRequest.
    """

    model = Declaration
    tracked_fields = [
        'date', 'desired_date', 'spec', 'work_type', 'children',
        'consent_full_time_group', 'desired_group_type', 'consent_dev_group',
        'consent_care_group', 'offer_other', 'best_privilege'
    ]

    def post_save(self, instance, created, **kwargs):

        origin_id = OrderRequest.objects.filter(
            declaration=instance,
            order_id__isnull=False
        ).values_list('order_id', flat=True).first()

        if self.has_changes(instance) and not created and origin_id:
            UpdateOrderRequestSendingTask().apply_async(
                kwargs={
                    'declaration_id': instance.id, 'order_id': origin_id
                }
            )


class DeclarationUnitChangesObserver(ChangesObserver):
    """
    Класс для отправки запроса UpdateOrderRequest при изменении полей
    желаемых учреждений заявления.
    """

    model = DeclarationUnit

    def post_save(self, instance, created, **kwargs):
        origin_id = OrderRequest.objects.filter(
            declaration=instance.declaration,
            order_id__isnull=False
        ).values_list('order_id', flat=True).first()

        if self.has_changes(instance) and not created and origin_id:
            UpdateOrderRequestSendingTask().apply_async(
                kwargs={
                    'declaration_id': instance.declaration.id,
                    'order_id': origin_id
                }
            )


class DirectStatusChangesObserver(ChangesObserver):
    """
    Класс для отправки запроса UpdateOrderRequest при изменении статуса
    направления.
    """

    model = DirectStatusLog

    def post_save(self, instance, created, **kwargs):
        origin_id = OrderRequest.objects.filter(
            declaration=instance.direct.declaration,
            order_id__isnull=False
        ).values_list('order_id', flat=True).first()

        if self.has_changes(instance) and not created and origin_id:
            UpdateOrderRequestSendingTask().apply_async(
                kwargs={
                    'declaration_id': instance.declaration.id,
                    'order_id': origin_id,
                    'direct_status_log': {
                        'id': instance.id,
                        'old_status': instance.old_status.code,
                        'status': instance.status.code,
                        'created_at': instance.created_at,
                    }
                }
            )
        declaration_message_id = DeclarationOriginMessageID.objects.filter(
            declaration=instance.direct.declaration).first()

        if created or not declaration_message_id:
            return

        message_id = declaration_message_id.message_id
        org_code, comment = get_code_and_comment(instance.direct)
        push_change_order_info_request(
            instance.direct.declaration.id, org_code, comment, message_id)


class ChildrenChangesObserver(ChangesObserver):
    """
    Класс для отправки запроса UpdateOrderRequest при изменении данных ребенка.
    """

    model = Children
    tracked_fields = [
        'surname', 'firstname', 'patronymic', 'date_of_birth',  'dul_series',
        'dul_number', 'dul_date', 'zags_act_number', 'zags_act_date',
        'zags_act_place', 'dul_type', 'reg_address_place',
        'reg_address_street', 'reg_address_house_guid', 'reg_address_full',
        'health_need', 'health_need_special_support', 'health_series',
        'health_number', 'health_need_start_date', 'health_issued_by',
        'health_need_expiration_date',
    ]

    def post_save(self, instance, created, **kwargs):
        for declaration in instance.declaration_set.all():
            origin_id = OrderRequest.objects.filter(
                declaration=declaration,
                order_id__isnull=False
            ).values_list('order_id', flat=True).first()

            if self.has_changes(instance) and not created and origin_id:
                UpdateOrderRequestSendingTask().apply_async(
                    kwargs={
                        'declaration_id': declaration.id,
                        'order_id': origin_id,
                    }
                )


class UnitObserver(ChangesObserver):
    """Класс для отправки запроса при изменении и удалении организации (ДОО).

    При наличии изменений определённых полей
    (Наименование, Адрес, Код ФИАС населенного пункта, Телефон, E-mail,
    Часы работы, ОКТМО, Адрес сайта)
    организации (ДОО) отправляется запрос CnsiRequest
    с вложением ClassifierDataUpdateRequest.

    При удаление организации (ДОО), при изменении статуса на ("Закрыто",
    "Ликвидировано", "Присоединена к другой организации")
    или проставлении признака "Не показывать на портале" отправляется
    запрос CnsiRequest с вложением ClassifierDataDeleteRequest.
    """

    model = Unit

    tracked_fields = [
        'name', 'address_full', 'address_place', 'telephone', 'email',
        'reception_time', 'octmo', 'site',
    ]

    unit_kind = UnitKind.DOU

    def pre_save(self, instance, context, **kwargs):
        """Сохраняется словарь с параметрами объекта до сохранения.

        Дополнительно сохраняет в инстансе атрибуты:
            1. esnsi_created - Признак того, что объект был создан в ЕСНСИ;
            2. esnsi_remove - Признак того, что объект возможно следует
                удалить в ЕСНСИ;
        """

        if not self.tracked_fields:
            self._pre_saved_instance_dict = {}

        elif instance.id is None:
            self._pre_saved_instance_dict = (
                self.get_instance_dict(instance, True))

            # Проставляет признак того, что новая организация еще
            # не создана в ЕСНСИ.
            instance.esnsi_created = False

        else:
            obj = self.model.objects.filter(id=instance.id).first()

            self._pre_saved_instance_dict = (
                self.get_instance_dict(obj) if obj else None)

            # Проставляет признак того, что организация была создана.
            # Если в БД у нее статус ("Закрыто", "Ликвидировано",
            # "Присоединена к другой организации") или проставлен чек-бокс
            # "Не показывать на портале".
            instance.esnsi_created = not (
                    obj.status in UnitStatus.ALL_CLOSED_STATUS
                    or obj.is_not_show_on_poral)

        # Если текущий статус организации:
        # "Закрыто", "Ликвидировано", "Присоединена к другой организации"
        # или проставлен чек-бокс "Не показывать на портале",
        # то проставляет признак возможного удаления.
        instance.esnsi_remove = (
                instance.status in UnitStatus.ALL_CLOSED_STATUS
                or instance.is_not_show_on_poral)

    def validate(self, instance):
        """Выполняет валидацию объекта.

        След. проверки:
            1. Организация является ДОО;
            2. Статус не "Закрыто", "Ликвидировано",
                "Присоединена к другой организации";
            3. Не проставлен признак "Не показывать на портале".

        :param instance: Организация
        :type instance: Unit

        :return: Признак успешности валидации
        :rtype: bool
        """

        return instance.kind_id == self.unit_kind and (
                instance.status not in UnitStatus.ALL_CLOSED_STATUS
                or not instance.is_not_show_on_poral)

    def post_save(self, instance, created, **kwargs):

        if instance.kind_id != self.unit_kind:
            return

        # Если организация уже была создана и нужно удалить,
        # то отправляет запрос на удаление.
        if instance.esnsi_created and instance.esnsi_remove:
            self.post_delete(
                instance=instance,
                context=self._get_context(instance),
                **kwargs)
            return

        # Если организация была создана и удалять не нужно,
        # то запросов не будет.
        if not instance.esnsi_created and instance.esnsi_remove:
            return

        if (
                (created or self.has_changes(instance))
                and self.validate(instance)
        ):
            ESNSIUnitTaskFactory(
                get_classifier_instance(UnitClassifier),
                UpdateClassifierRequest.UPDATE_CLASSIFIERS,
                instance
            ).set_task()

    def pre_delete(self, instance, context, **kwargs):

        instance.octmo = instance.octmo or instance.get_octmo()
        if self.validate(instance) and instance.octmo:
            ESNSIUnitTaskFactory(
                get_classifier_instance(UnitClassifier),
                UpdateClassifierRequest.DELETE_CLASSIFIERS,
                instance
            ).set_task()


class MaxDooObserver(ChangesObserver):
    """Класс для отправки запроса при изменении и удалении организации (МО).

    При наличии изменений определённых полей
    (Наименование, ОКТМО, Максимальное количество желаемых ДОО в заявке)
    организации (МО) отправляется запрос CnsiRequest
    с вложением ClassifierDataUpdateRequest.

    При удаление организации (МО) отправляется запрос CnsiRequest
    с вложением ClassifierDataDeleteRequest.
    """

    model = Unit

    tracked_fields = ['name', 'octmo', 'max_desired_dou']

    unit_kind = UnitKind.MO

    def validate(self, instance):
        """Выполняет валидацию объекта.

        Проверяет, что организация является МО.

        :param instance: Организация
        :type instance: Unit

        :return: Признак успешности валидации
        :rtype: bool
        """

        return instance.kind_id == self.unit_kind

    def post_save(self, instance, created, **kwargs):

        if (
                (created or self.has_changes(instance))
                and self.validate(instance)
        ):

            ESNSIUnitTaskFactory(
                get_classifier_instance(MaxDooClassifier),
                UpdateClassifierRequest.UPDATE_CLASSIFIERS,
                instance
            ).set_task()

    def post_delete(self, instance, context, **kwargs):

        if self.validate(instance):

            ESNSIUnitTaskFactory(
                get_classifier_instance(MaxDooClassifier),
                UpdateClassifierRequest.DELETE_CLASSIFIERS,
                instance
            ).set_task()


declaration_status_changes_observer = DeclarationStatusChangesObserver()
direct_status_changes_observer = DirectStatusChangesObserver()
change_declaration_observer = ChangeDeclarationObserver()
unit_observer = UnitObserver()
max_doo_observer = MaxDooObserver()
declaration_fields_changes_observer = DeclarationFieldsChangesObserver()
declaration_unit_changes_observer = DeclarationUnitChangesObserver()
children_changes_observer = ChildrenChangesObserver()
