from abc import ABC
from abc import abstractmethod

from aio_client.base import RequestTypeEnum
from aio_client.consumer.api import push_request
from aio_client.consumer.models import PostConsumerRequest
from django.db.models import OuterRef
from django.db.models import Subquery
from educommon.ws_log.models import SmevLog
from educommon.ws_log.models import SmevSourceEnum
from m3_fias.utils import get_address_object
from m3_fias.utils import get_address_object_name
from m3_fias.utils import get_house

from concentrator.smev3_v321 import settings
from concentrator.smev3_v321.base.utils import get_adaptation_group
from concentrator.smev3_v321.base.utils import get_adaptation_group_type
from concentrator.smev3_v321.base.utils import get_identity_doc_name
from concentrator.smev3_v321.base.utils import get_language
from concentrator.smev3_v321.base.utils import get_medical_report
from concentrator.smev3_v321.base.utils import get_schedule
from concentrator.smev3_v321.base.utils import is_cancel_allowed
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.models import ApplicantAnswer
from concentrator.smev3_v321.models import OrderRequest
from concentrator.smev3_v321.models import UpdateOrderRequest
from concentrator.smev3_v321.service_types import kinder_order
from kinder.core.declaration.models import Declaration
from kinder.core.declaration.models import DeclarationPrivilege
from kinder.core.declaration_status.models import DeclarationStatusTransfer
from kinder.core.dict.models import DULTypeEnumerate
from kinder.core.direct.models import DRS
from kinder.core.direct.models import DirectRefusalReason
from kinder.core.direct.models import RefusalReasonInitiatorType
from kinder.core.utils.fias import name_from_guid

from .constants import DECLARATION_FORMATION
from .constants import DELEGATE_REFUSED
from .constants import DIRECTED_TO_DOO
from .constants import NEED_NOT_CONFIRMED
from .constants import WAITING_FOR_CONTRACT
from .constants import CreateOrderStatusMapper
from .constants import UpdateOrderStatusMapper


class OrderRequestBaseHelper(ABC):
    """Базовый класс отправки запросов OrderRequest."""

    service_name = None

    def __init__(self, declaration_id, order_id, direct=None,
                 delegate_id=None):
        self.declaration = Declaration.objects.get(id=declaration_id)
        self.order_id = order_id
        self.delegate_id = delegate_id

        self.direct = direct or self.declaration.direct_set.all().first()
        self.epgu_status_mapper = CreateOrderStatusMapper(
            self.declaration, self.direct)

    @classmethod
    def log(cls, request_obj, result=None, logging_data=None):
        """
        Логирование запроса СМЭВ 3.

        :param request_obj: Сообщение
        :param result: Результат
        :param logging_data: Дополнительные параметры
        """

        logging_data = logging_data or {}

        data = {
            'service_address':
                RequestTypeEnum.get_url(RequestTypeEnum.CS_POST),
            'method_name': cls.service_name,
            'method_verbose_name':
                f'{cls.service_name} (Формирование и отправка запроса '
                f'OrderRequest "Передача данных заявления в ЛК ЕПГУ".)',
            'direction': SmevLog.OUTGOING,
            'interaction_type': SmevLog.IS_SMEV,
            'source': SmevSourceEnum.CONCENTRATOR,
            'request': getattr(request_obj, 'body', None),
            'result': result,
            **logging_data
        }

        return SmevLog(**data).save()

    @abstractmethod
    def get_request(self):
        """Формирование запроса"""

    @abstractmethod
    def send_request(self):
        """Отправка запроса"""


class OrderRequestHelper(OrderRequestBaseHelper):
    """
    Формирование и отправка запроса OrderRequest
    "Передача данных заявления в ЛК ЕПГУ".

    """

    service_name = 'OrderRequest'

    def get_status_history_data(self):
        """Получение данных по истории статусов заявки"""

        history_data = []

        decl_status_code = self.declaration.status.code

        if not self.declaration.declarationstatuslog_set.exists():

            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[{'status__code': decl_status_code}]
            )

            return kinder_order.statusHistoryType(
                statusCode=kinder_order.statusCodeType(
                    orgCode=epgu_status_code),
                statusDate=self.declaration.date,
                statusComment=epgu_comment,
                cancelAllowed=is_cancel_allowed(self.declaration.status)
            )

        auto_change_days = Subquery(
            DeclarationStatusTransfer.objects.filter(
                from_status=OuterRef('old_status'),
                to_status=OuterRef('status'),
                auto_change_days__gt=0
            ).values('auto_change_days')[:1]
        )

        decl_status_log = self.declaration.declarationstatuslog_set.annotate(
            auto_change_days=auto_change_days
        ).values(
            'id', 'old_status__code', 'status__code', 'datetime',
            'comment', 'auto_change_days'
        ).all()

        for status_history in decl_status_log:
            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[status_history]
            )

            history_data.append(kinder_order.statusHistoryType(
                statusCode=kinder_order.statusCodeType(
                    orgCode=epgu_status_code
                ),
                statusDate=status_history['datetime'],
                statusComment=epgu_comment,
                cancelAllowed=is_cancel_allowed(self.declaration.status)
            ))

        return kinder_order.statusHistoryListType(history_data)

    @staticmethod
    def get_child_doc_data(children):
        """Получение данных по ребенку"""

        if children.dul_type != DULTypeEnumerate.INT_SVID:
            return {
                'ChildBirthDocRF': kinder_order.ChildBirthDocRFType(
                    ChildBirthDocSeries=children.dul_series,
                    ChildBirthDocNumber=children.dul_number,
                    ChildBirthDocIssueDate=children.dul_date,
                    ChildBirthDocActNumber=children.zags_act_number,
                    ChildBirthDocActDate=children.zags_act_date,
                    ChildBirthDocIssued=children.zags_act_place
                )
            }
        else:
            return {
                'ChildBirthDocForeign': kinder_order.ChildBirthDocForeignType(
                    ChildBirthDocName=DULTypeEnumerate.get_value_by_id(
                        children.dul_type),
                    ChildBirthDocSeries=children.dul_series,
                    ChildBirthDocNumber=children.dul_number,
                    ChildBirthDocIssueDate=children.dul_date,
                    ChildBirthDocIssued=children.zags_act_place
                )
            }

    def get_address_data(self, children):
        """Получение данных для адреса"""

        address_dict = self.get_address_dict(
            children.reg_address_place,
            children.reg_address_street,
            children.reg_address_house_guid
        )

        return kinder_order.AddressType(
            FullAddress=children.reg_address_full,
            Index=address_dict.get('postal_code', ''),
            Region=kinder_order.DataElementType(
                code=children.reg_address_place,
                valueOf_=address_dict.get('place_name', '')
            ),
            Area=kinder_order.DataElementType(
                code=children.reg_address_place,
                valueOf_=name_from_guid(children.reg_address_place),
            ),
            City=kinder_order.DataElementType(
                code=children.reg_address_place,
                valueOf_=name_from_guid(children.reg_address_place),
            ),
            Place=kinder_order.DataElementType(
                code=children.reg_address_place,
                valueOf_=name_from_guid(children.reg_address_place),
            ),
            Street=kinder_order.DataElementType(
                code=children.reg_address_street,
                valueOf_=address_dict.get('street_name', '')
            ),
            House=kinder_order.DataElementType(
                code=children.reg_address_house,
                valueOf_=address_dict.get('house_number', '')
            ),
            Building1=children.reg_address_corps,
            Building2=address_dict.get('structure_number', ''),
            Apartment=children.reg_address_flat
        )

    @staticmethod
    def get_address_dict(place_guid, street_guid, house_guid):
        """Формирование словаря с данными для адреса"""

        address = {}
        if house_guid and street_guid:
            house_obj = get_house(house_guid, street_guid)
            if house_obj:
                address['house_number'] = house_obj.house_number
                address['building_number'] = house_obj.building_number
                address['postal_code'] = house_obj.postal_code
                address['structure_number'] = house_obj.structure_number
        if street_guid:
            street_obj = get_address_object(street_guid)
            if street_obj:
                address['street_name'] = get_address_object_name(street_obj)
                if not address.get('postal_code'):
                    address['postal_code'] = street_obj.postal_code
        if place_guid:
            place_obj = get_address_object(place_guid)
            if place_obj:
                address['place_name'] = place_obj.full_name
                if not address.get('postal_code'):
                    address['postal_code'] = place_obj.postal_code

        return address

    def get_edu_organizations_data(self):
        """Получение данных по желаеммым ДОО в заявке"""

        organizations = []

        for decl_unit in self.declaration.declarationunit_set.all():
            organizations.append(kinder_order.EduOrganizationType(
                code=decl_unit.unit_id,
                PriorityNumber=decl_unit.ord,
                valueOf_=decl_unit.unit.name
            ))

        return organizations

    @staticmethod
    def get_delegatecontingent_data(delegate):
        """Получение данных представителя из Контингента"""

        if hasattr(delegate, 'delegatecontingent'):
            return kinder_order.OtherRepresentativeType(
                OtherRepresentativeDocName=delegate.delegatecontingent.doc_type,
                OtherRepresentativeDocSeries=delegate.delegatecontingent.series,
                OtherRepresentativeDocNumber=delegate.delegatecontingent.number,
                OtherRepresentativeDocDate=delegate.delegatecontingent.date_issue,
                OtherRepresentativeDocIssued=delegate.delegatecontingent.issued_by,
            )

    @staticmethod
    def get_medical_data(children):
        """
        Получение данных по "Реквизитам документа, подтверждающего группу
        коменсирующей направленности".
        """

        if (
                children.health_number and children.health_need_start_date
                and children.health_issued_by
        ):
            return kinder_order.MedicalReportWithoutFilesType(
                DocName=kinder_order.DataElementType(
                    **get_medical_report(children)),
                DocSeries=children.health_series,
                DocNumber=children.health_number,
                DocIssueDate=children.health_need_start_date,
                DocIssued=children.health_issued_by,
                DocExpirationDate=children.health_need_expiration_date)

    def get_application_data(self):
        """Получение всех данных по заявке"""

        if self.delegate_id:
            delegate = self.declaration.children.childrendelegate_set.get(
                delegate__id=self.delegate_id
            ).delegate
        else:
            delegate = (
                self.declaration.children.childrendelegate_set.first().delegate
            )
        children = self.declaration.children

        delegatecontingent_data = self.get_delegatecontingent_data(delegate)

        application_data = kinder_order.ApplicationType(
            PersonIdentityDocInfo=kinder_order.PersonIdentityDocInfoType(
                IdentityDocName=kinder_order.DataElementType(
                    **get_identity_doc_name(delegate.dul_type)),
                IdentityDocSeries=delegate.dul_series,
                IdentityDocNumber=delegate.dul_number,
                IdentityDocIssueDate=delegate.dul_date,
                IdentityDocIssueCode=delegate.dul_place,
                IdentityDocIssued=delegate.dul_issued_by,
            ),
            ChildInfo=kinder_order.ChildInfoType(
                ChildSurname=children.surname,
                ChildName=children.firstname,
                ChildMiddleName=children.patronymic,
                ChildBirthDate=children.date_of_birth,
                **self.get_child_doc_data(children),
            ),
            Address=self.get_address_data(children),
            EntryParams=kinder_order.EntryParamsType(
                EntryDate=self.declaration.desired_date,
                Language=kinder_order.DataElementType(
                    **get_language(self.declaration.spec)),
                Schedule=kinder_order.DataElementType(
                    **get_schedule(self.declaration.work_type)),
                AgreementOnFullDayGroup=self.declaration.consent_full_time_group
            ),
            AdaptationProgram=kinder_order.AdaptationProgramType(
                AdaptationGroup=kinder_order.DataElementType(
                    **get_adaptation_group(self.declaration.desired_group_type)
                ),
                AdaptationGroupType=kinder_order.DataElementType(
                    **get_adaptation_group_type(
                        children.health_need, self.declaration.work_type.code)),
                AgreementOnGeneralGroup=self.declaration.consent_dev_group,
                AgreementOnCareGroup=self.declaration.consent_care_group,
                NeedSpecialCareConditions=children.health_need_special_support,
            ),
            MedicalReport=self.get_medical_data(children),
            EduOrganizations=kinder_order.EduOrganizationsType(
                EduOrganization=self.get_edu_organizations_data(),
                AllowOfferOther=self.declaration.offer_other,
            ),
        )

        person_info = kinder_order.PersonInfoType(
                PersonSurname=delegate.surname,
                PersonName=delegate.firstname,
                PersonMiddleName=delegate.patronymic,
                PersonPhone=delegate.phones or delegate.phone_for_sms or '-',
                PersonEmail=delegate.email or '-',
                Parents='true' if delegate.type else '',
            )
        if delegatecontingent_data:
            person_info.set_OtherRepresentative(
                delegatecontingent_data
            )

        application_data.set_PersonInfo(person_info)

        for declaration_unit in self.declaration.declarationunit_set.all():
            if declaration_unit.sibling:
                application_data.add_BrotherSisterInfo(
                    kinder_order.BrotherSisterInfoType(
                        ChildSurname=declaration_unit.sibling.surname,
                        ChildName=declaration_unit.sibling.firstname,
                        ChildMiddleName=declaration_unit.sibling.patronymic,
                        EduOrganization=kinder_order.DataElementType(
                            code=declaration_unit.unit_id,
                            valueOf_=declaration_unit.unit.name,
                        ),
                    )
                )

        declaration_privilege = DeclarationPrivilege.objects.filter(
            privilege=self.declaration.best_privilege,
            declaration=self.declaration
        ).first()

        if declaration_privilege:

            end_date_info = {}
            if declaration_privilege._privilege_end_date:
                end_date_info['DocExpirationDate'] = (
                    declaration_privilege._privilege_end_date)

            doc_info = kinder_order.DocInfoType(
                DocIssueDate=declaration_privilege.doc_date,
                DocIssued=declaration_privilege.doc_issued_by,
                **end_date_info
            )

            application_data.set_BenefitInfo(
                kinder_order.BenefitInfoWithoutFilesType(
                    BenefitCategory=kinder_order.DataElementType(
                        code=declaration_privilege.privilege.esnsi_code,
                        valueOf_=declaration_privilege.privilege.name,
                    ),
                    BenefitDocInfo=doc_info
                )
            )

        return application_data

    def get_request(self):
        request = kinder_order.OrderRequestType(
            env=settings.ORDER_REQUEST_SERVICE_ENV,
            CreateOrderRequest=kinder_order.CreateOrderRequestType(
                orderId_InfoRequest=self.order_id,
                requestDate=self.declaration.date,
                statusHistoryList=self.get_status_history_data(),
                application=self.get_application_data(),)
        )

        request = render_type2xml(
            request,
            name_type=self.service_name
        )

        return request

    def send_request(self):
        request = self.get_request()

        order_request, _ = OrderRequest.objects.get_or_create(
            declaration=self.declaration,
            defaults={'request_order_id': self.order_id}
        )
        order_request.generate_message_id()

        request_obj = PostConsumerRequest(
            message_id=str(order_request.message_id),
            body=request,
            message_type=self.service_name
        )

        push_request(request_obj)

        return request_obj


class UpdateRequestHelper(OrderRequestHelper):
    """
    Формирование и отправка информации об изменении статуса заявления
    по ВС "Передача заявлений на запись в дошкольную организацию".
    Запрос OrderRequest с блоком UpdateOrderRequest.

    """

    service_name = 'UpdateOrderRequest'

    # Именованные аргументы с которыми может вызыватся асинхронная задача и
    # соотвественно данный хелпер. Используются в маппинге статусов для
    # формирования комментария ЕПГУ
    init_kwargs = (
        'decl_status_changed',
        'direct_status_log',
        'event',
        'declaration_changes_rows',
        'reject_changes_comment'
    )

    def __init__(self, declaration_id, order_id, direct=None, **kwargs):
        super().__init__(declaration_id, order_id, direct)

        for kwarg in kwargs:
            assert kwarg in self.init_kwargs

        self.decl_status_changed = kwargs.get('decl_status_changed', False)
        self.direct_status_log = kwargs.get('direct_status_log', None)
        self.event = kwargs.get('event', None)

        self.__dict__.update(kwargs)

        self.epgu_status_mapper = UpdateOrderStatusMapper(
            self.declaration, direct, **kwargs)

    def get_status_history_data(self):
        """
        Получение данных по статусу заявки (код и комментарий для ЕПГУ).

        """

        auto_change_days = Subquery(
            DeclarationStatusTransfer.objects.filter(
                from_status=OuterRef('old_status'),
                to_status=OuterRef('status'),
                auto_change_days__gt=0
            ).values('auto_change_days')[:1]
        )

        current_decl_status_code = self.declaration.status.code
        status_date = self.declaration.date

        # Произошло изменение статуса заявления,
        # достаем текущий установленный статус
        if self.decl_status_changed:
            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[
                    {'status__code': current_decl_status_code}
                ]
            )
        # Произошло изменение статуса направления
        elif self.direct_status_log:
            new_direct_status = self.direct_status_log['status']

            if new_direct_status == DRS.REGISTER:
                has_applicant_answer = ApplicantAnswer.objects.filter(
                    direct=self.direct
                ).values_list('answer', flat=True).first()
                if has_applicant_answer:
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {'status__code': DIRECTED_TO_DOO}
                        ]
                    )
                else:
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {'status__code': DECLARATION_FORMATION}
                        ]
                    )
            elif new_direct_status in (DRS.REJECT, DRS.REFUSED):
                initiator_type = DirectRefusalReason.objects.filter(
                    direct_status_log_id=self.direct_status_log['id']
                ).values_list('reason__initiator_type', 'reason__name').first()

                if (initiator_type['reason__initiator_type'] ==
                        RefusalReasonInitiatorType.EMPLOYEE):
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {
                                'status__code': NEED_NOT_CONFIRMED,
                                'reason': initiator_type['reason__name'],
                                'created_at': self.direct_status_log['created_at'],
                                'old_status': self.direct_status_log['old_status'],
                            }
                        ]
                    )
                elif initiator_type == RefusalReasonInitiatorType.PARENT:
                    epgu_status_code, epgu_comment = (
                        self.epgu_status_mapper[
                            {'status__code': DELEGATE_REFUSED}
                        ]
                    )
            elif new_direct_status == DRS.DOGOVOR:
                epgu_status_code, epgu_comment = (
                    self.epgu_status_mapper[
                        {'status__code': WAITING_FOR_CONTRACT}
                    ]
                )
        # Произошло событие "Отправка в ЭДС запроса ApplicationRequest" или
        # "В ЭДС пришло изменение и изменения отклонены"
        # (Заявление - вкладка "Изменения с ЕПГУ" - выбрать запись - Отказать)
        elif self.event:
            epgu_status_code, epgu_comment = (
                self.epgu_status_mapper[
                    {'status__code': self.event}
                ]
            )
        # Произошло событие изменение полей заявления (кроме статуса).
        # Достаем последний установленный статус из лога статусов,
        # если он пуст, то берем текуший статус заявки
        else:
            decl_status_log = self.declaration.declarationstatuslog_set.annotate(
                auto_change_days=auto_change_days
            ).values(
                'id', 'old_status__code', 'status__code', 'datetime',
                'comment', 'auto_change_days'
            ).order_by('id').last()

            if decl_status_log:
                epgu_status_code, epgu_comment = (
                    self.epgu_status_mapper[decl_status_log]
                )
                status_date = decl_status_log['datetime']
            else:
                epgu_status_code, epgu_comment = (
                    self.epgu_status_mapper[
                        {'status__code': current_decl_status_code}
                    ]
                )

        return kinder_order.statusHistoryType(
            statusCode=kinder_order.statusCodeType(
                orgCode=epgu_status_code),
            statusDate=status_date,
            statusComment=epgu_comment,
            cancelAllowed=is_cancel_allowed(self.declaration.status)
        )

    def get_request(self):
        """Формирование запроса"""

        if self.decl_status_changed or self.direct_status_log or self.event:
            request = kinder_order.OrderRequestType(
                env=settings.ORDER_REQUEST_SERVICE_ENV,
                UpdateOrderRequest=kinder_order.UpdateOrderRequestType(
                    orderId=int(self.order_id),
                    statusHistoryList=self.get_status_history_data(),
                )
            )
        else:
            request = kinder_order.OrderRequestType(
                env=settings.ORDER_REQUEST_SERVICE_ENV,
                UpdateOrderRequest=kinder_order.UpdateOrderRequestType(
                    orderId=int(self.order_id),
                    statusHistoryList=self.get_status_history_data(),
                    application=self.get_application_data(),
                )
            )

        request = render_type2xml(
            request,
            name_type=self.service_name
        )

        return request

    def send_request(self):
        """Отправка запроса"""

        request = self.get_request()

        update_request = UpdateOrderRequest.objects.create(
            declaration=self.declaration,
        )
        update_request.generate_message_id()

        request_obj = PostConsumerRequest(
            message_id=str(update_request.message_id),
            body=request,
            message_type=self.service_name
        )

        push_request(request_obj)

        return request_obj
